#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

"""        Read a file and call a handler for each line. Update progress bar if present.  """
import logging
import os


class FileReader:

    def __init__(self, directory: str, filename: str, update_progress):
        """
        Read a file and call a handler for each line.  Update progress bar if present
        #Args:
            directory:
            filename:
            progress_bar: Progress Bar or None
        """
        self.logger = logging.getLogger(__name__)
        self.directory: str = directory
        self.update_progress = update_progress
        self.fname: str = filename
        self.cache_changed = False
        self.count = 0

    def read(self) -> bool:
        """
        Read a file and call a handler for each line.  Update progress bar
        :return: Error
        """
        line_num = 0
        file_pos = 0

        path = os.path.join(self.directory, self.fname)
        self.logger.info(f"Reading file {path}")
        if os.path.exists(path):
            fsize = os.path.getsize(path)
            with open(path, 'r', newline="", encoding='utf-8', errors='replace') as file:
                for row in file:
                    line_num += 1
                    file_pos += len(row)
                    self.handle_line(line_num, row)
                    if line_num % 80000 == 1:
                        # Periodically update progress
                        prog = file_pos * 100 / fsize
                        self.progress(f"2) Loading {self.fname} {prog:.0f}%", prog)

            self.cache_changed = True
            self.progress("", 100)
            self.logger.info(f'Added {self.count} items')
            return False
        else:
            self.logger.error(f'Unable to open {path}')
            return True

    def cancel(self):
        # User requested cancel of file read
        pass

    def handle_line(self, line_num: int, row: str) -> int:
        # Handle the line we read
        pass

    def progress(self, msg, val):
        """ Update progress bar if there is one """
        if self.update_progress is not None:
            self.update_progress(val, msg)

        # If we're past 80% log item as info, otherwise log as debug
        if val > 90 or val < 10 or (val > 40 and val < 50):
            self.logger.info(f'{val:.1f}%  {msg}')
        else:
            self.logger.debug(f'{val:.1f}%  {msg}')




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""
Provide functions to normalize text strings by converting to lowercase, removing noisewords.   
This is used by the lookup functions and the database build functions and match scoring   
noise_words is a list of replacements only used for match scoring   
phrase_cleanup is a list of replacements for db build, lookup and match scoring   
"""
import re
import sys

import unidecode

from geodata import GeodataBuild, Loc, GeoUtil

# Todo -  make these all list driven

noise_words = [
    # Noise words - Replacements done for match scoring
    (r', ', ','),
    (r"normandy american ", 'normandie american '),
    (r'nouveau brunswick', ' '),
    (r'westphalia', 'westfalen'),
    (r"l'", ''),
    (r'citta metropolitana di ', ' '),
    (r'kommune', ''),
    (r"politischer bezirk ", ' '),
    (r'erry', 'ury'),
    (r'ery', 'ury'),
    (r'borg', 'burg'),
    (r'bourg', 'burg'),
    (r'urgh', 'urg'),
    (r'mound', 'mund'),
    (r'ourne', 'orn'),
    (r'ney', 'ny'),
    (r' de ', ' '),
    (r' di ', ' '),
    (r' du ', ' '),
    (r' of ', ' '),
    (r' departement', ' '),
    (r'royal borough of windsor and maidenhead', 'berkshire'),
    (r'regional municipality', 'county'),
    ]

phrase_cleanup = [
    # Phrase cleanup - replacements that are always applied (for database build, lookup, and match scoring)  
    ('r\.k\. |r k ', 'roman catholic '),
    ('rooms katholieke ', 'roman catholic '),
    ('sveti |saints |sainte |sint |saint |sankt |st\. ', 'st '),  # Normalize Saint to St
    (r' co\.', ' county'),  # Normalize County
    (r'united states of america', 'usa'),  # Normalize to USA   begraafplaats
    (r'united states', 'usa'),  # Normalize to USA
    (r'town of ', ''),  # - remove town of
    (r'city of ', ''),  # - remove city of
    (r'near ', ' '),  # - remove near
    (r'cimetiere', 'cemetery'),  # cimetière
    (r'begraafplaats', 'cemetery'),  # 
    ('  +', ' '),  # Strip multiple space
    ('county of ([^,]+)', r'\g<1> county'),  # Normalize 'Township of X' to 'X Township'
    ('township of ([^,]+)', r'\g<1> township'),  # Normalize 'Township of X' to 'X Township'
    ('cathedral of ([^,]+)', r'\g<1> cathedral'),  # Normalize 'Township of X' to 'X Township'
    ('palace of ([^,]+)', r'\g<1> palace'),  # Normalize 'Township of X' to 'X Township'
    ('castle of ([^,]+)', r'\g<1> castle'),  # Normalize 'Township of X' to 'X Township'
    (r',castle', ' castle'),  # -  remove extra comma
    (r',palace', ' palace'),  # -  remove extra comma
    (r"'(\w{2,})'",r"\g<1>"), # remove single quotes around word
    ]


def normalize_for_scoring(text: str, iso: str) -> str:
    """
        Normalize the title we use to determine how close a match we got. 
        See normalize() for details  
        Also remove noise words such as City Of

    #Args:
        text: text to normalize
        iso: ISO country code

    #Returns:

    """
    text = normalize(text=text, remove_commas=False)
    text = _remove_noise_words(text)
    return text


def normalize(text: str, remove_commas: bool) -> str:
    """
    Normalize text - Convert from UTF-8 to lowercase ascii.  
    Remove commas if parameter set.   
    Remove all non alphanumeric except $ and *  
    Then call _phrase_normalize() which normalizes common phrases with multiple spellings, such as saint to st   
    #Args:   
        text:  Text to normalize   
        remove_commas:   True if commas should be removed   

    #Returns:   
        Normalized text

    """

    # Convert UT8 to ascii
    text = unidecode.unidecode(text)
    text = str(text).lower()

    # remove all non alphanumeric except $ and * and comma(if flag set)
    if remove_commas:
        text = re.sub(r"[^a-z0-9 $*']+", " ", text)
    else:
        text = re.sub(r"[^a-z0-9 $*,']+", " ", text)

    text = _phrase_normalize(text)
    return text.strip(' ')


def _phrase_normalize(text: str) -> str:
    """ Strip spaces and normalize spelling for items such as Saint and County """
    # Replacement patterns to clean up entries

    if 'amt' not in text:
        text = re.sub(r'^mt ', 'mount ', text)

    for pattern, replace in phrase_cleanup:
        text = re.sub(pattern, replace, text)

    return text


def _remove_noise_words(text: str):
    # Calculate score with noise words removed    
    for pattern, replace in noise_words:
        text = re.sub(pattern, replace, text)
    return text


def remove_aliase(input_words, res_words) -> (str, str):
    if "middlesex" in input_words and "greater london" in res_words:
        input_words = re.sub('middlesex', '', input_words)
        res_words = re.sub('greater london', '', res_words)
    return input_words, res_words


alias_list = {
    'norge'               : ('norway', '', 'ADM0'),
    'sverige'             : ('sweden', '', 'ADM0'),
    'osterreich'          : ('austria', '', 'ADM0'),
    'belgie'              : ('belgium', '', 'ADM0'),
    'brasil'              : ('brazil', '', 'ADM0'),
    'danmark'             : ('denmark', '', 'ADM0'),
    'eire'                : ('ireland', '', 'ADM0'),
    'magyarorszag'        : ('hungary', '', 'ADM0'),
    'italia'              : ('italy', '', 'ADM0'),
    'espana'              : ('spain', '', 'ADM0'),
    'deutschland'         : ('germany', '', 'ADM0'),
    'prussia'             : ('germany', '', 'ADM0'),
    'suisse'              : ('switzerland', '', 'ADM0'),
    'schweiz'             : ('switzerland', '', 'ADM0'),

    'bayern'              : ('bavaria', 'de', 'ADM1'),
    'westphalia'          : ('westfalen', 'de', 'ADM1'),

    'normandy'            : ('normandie', 'fr', 'ADM1'),
    'brittany'            : ('bretagne', 'fr', 'ADM1'),
    'burgundy'            : ('bourgogne franche comte', 'fr', 'ADM1'),
    'franche comte'       : ('bourgogne franche comte', 'fr', 'ADM1'),
    'aquitaine'           : ('nouvelle aquitaine', 'fr', 'ADM1'),
    'limousin'            : ('nouvelle aquitaine', 'fr', 'ADM1'),
    'poitou charentes'    : ('nouvelle aquitaine', 'fr', 'ADM1'),
    'alsace'              : ('grand est', 'fr', 'ADM1'),
    'champagne ardenne'   : ('grand est', 'fr', 'ADM1'),
    'lorraine'            : ('grand est', 'fr', 'ADM1'),
    'languedoc roussillon': ('occitanie', 'fr', 'ADM1'),
    'nord pas de calais'  : ('hauts de france', 'fr', 'ADM1'),
    'picardy'             : ('hauts de france', 'fr', 'ADM1'),
    'auvergne'            : ('auvergne rhone alpes', 'fr', 'ADM1'),
    'rhone alpes'         : ('auvergne rhone alpes', 'fr', 'ADM1'),

    'breconshire'         : ('sir powys', 'gb', 'ADM2'),
    }

ALIAS_FEAT = 2
ALIAS_ISO = 1
ALIAS_NAME = 0


def add_aliases_to_db(geo_build: GeodataBuild):
    #  Add alias names to DB
    for ky in alias_list:
        add_alias_to_db(ky, geo_build)


def add_alias_to_db(ky: str, geo_build: GeodataBuild):
    alias_row = alias_list.get(ky)
    place = Loc.Loc()
    place.country_iso = alias_row[ALIAS_ISO].lower()
    place.city = alias_row[ALIAS_NAME]
    place.feature = alias_row[ALIAS_FEAT]
    place.place_type = Loc.PlaceType.CITY

    # Lookup main entry and get GEOID
    geo_build.geodb.s.lookup_place(place)
    if len(place.georow_list) > 0:
        if len(place.georow_list[0]) > 0:
            geo_row = list(place.georow_list[0][0:GeoUtil.Entry.SDX+1])
            geo_build.update_geo_row_name(geo_row=geo_row, name=ky)
            geo_tuple = tuple(geo_row)
            geo_build.insert(geo_tuple=geo_tuple, feat_code=alias_row[ALIAS_FEAT])


def deb(msg=None):
    # Display debug message with line number
    print(f"Debug {sys._getframe().f_back.f_lineno}: {msg if msg is not None else ''}")


def admin1_normalize(admin1_name: str, iso):
    """ Normalize historic or colloquial Admin1 names to current geoname standard """
    admin1_name = normalize(admin1_name, False)
    if iso == 'de':
        admin1_name = re.sub(r'bayern', 'bavaria', admin1_name)
        admin1_name = re.sub(r'westphalia', 'westfalen', admin1_name)
    elif iso == 'fr':
        admin1_name = re.sub(r'normandy', 'normandie', admin1_name)
        admin1_name = re.sub(r'brittany', 'bretagne', admin1_name)
        admin1_name = re.sub(r'burgundy', 'bourgogne franche comte', admin1_name)
        admin1_name = re.sub(r'franche comte', 'bourgogne franche comte', admin1_name)
        admin1_name = re.sub(r'aquitaine', 'nouvelle aquitaine', admin1_name)
        admin1_name = re.sub(r'limousin', 'nouvelle aquitaine', admin1_name)
        admin1_name = re.sub(r'poitou charentes', 'nouvelle aquitaine', admin1_name)
        admin1_name = re.sub(r'alsace', 'grand est', admin1_name)
        admin1_name = re.sub(r'champagne ardenne', 'grand est', admin1_name)
        admin1_name = re.sub(r'lorraine', 'grand est', admin1_name)
        admin1_name = re.sub(r'languedoc roussillon', 'occitanie', admin1_name)
        admin1_name = re.sub(r'midi pyrenees', 'occitanie', admin1_name)
        admin1_name = re.sub(r'nord pas de calais', 'hauts de france', admin1_name)
        admin1_name = re.sub(r'picardy', 'hauts de france', admin1_name)
        admin1_name = re.sub(r'auvergne', 'auvergne rhone alpes', admin1_name)
        admin1_name = re.sub(r'rhone alpes', 'auvergne rhone alpes', admin1_name)
    elif iso == 'ca':
        admin1_name = re.sub(r'brunswick', 'brunswick*', admin1_name)

    return admin1_name


def admin2_normalize(admin2_name: str, iso) -> (str, bool):
    """
        Normalize historic or colloquial Admin2 names to standard

    Args:
        admin2_name: 
        iso: 

    Returns: TUPLE (result, modified) - result is new string, modified - True if modified

    """
    admin2_name = normalize(admin2_name, False)

    mod = False

    if iso == 'gb':
        admin2_name = re.sub(r'breconshire', 'sir powys', admin2_name)
        mod = True

    return admin2_name, mod


def country_normalize(country_name) -> (str, bool):
    """
    normalize local language Country name to standardized English country name for lookups
    :param country_name:
    :return: (result, modified)
    result - new string
    modified - True if modified
    """
    country_name = re.sub(r'\.', '', country_name)  # remove .

    natural_names = {
        'norge'       : 'norway',
        'sverige'     : 'sweden',
        'osterreich'  : 'austria',
        'belgie'      : 'belgium',
        'brasil'      : 'brazil',
        'danmark'     : 'denmark',
        'eire'        : 'ireland',
        'magyarorszag': 'hungary',
        'italia'      : 'italy',
        'espana'      : 'spain',
        'deutschland' : 'germany',
        'prussia'     : 'germany',
        'suisse'      : 'switzerland',
        'schweiz'     : 'switzerland',

        }
    if natural_names.get(country_name):
        country_name = natural_names.get(country_name)
        return country_name.strip(' '), True
    else:
        return country_name.strip(' '), False

def normalize_features(feature, name, pop:int):
    # Set city feature based on population and cleanup features for RUIN and HSTS
    feat = feature
    
    # Create new Feature codes based on city population
    if pop > 1000000 and 'PP' in feature:
        feat = 'PP1M'
    elif pop > 100000 and 'PP' in feature:
        feat = 'P1HK'
    elif pop > 10000 and 'PP' in feature:
        feat = 'P10K'
    
    # Set feature type for Abbey/Priory, Castle, and Church if RUIN or HSTS
    if feature == 'HSTS' or feature == 'RUIN':
        if 'abbey' in name or 'priory' in name:
            # Change feature to MSTY plus first letter of original feature code (H or R)
            feat = 'MSTY' + feature[0]
        elif 'priory' in name:
            # Change feature to MSTY plus first letter of original feature code
            feat = 'MSTY' + feature[0]
        elif 'castle' in name:
            # Change feature to MSTY plus first letter of original feature code
            feat = 'CSTL' + feature[0]
        elif 'church' in name:
            # Change feature to MSTY plus first letter of original feature code
            feat = 'CH_' + feature[0]
    return feat



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""Calculate a heuristic score for how well a result place name matches a target place name."""
import copy
import logging

# import python-Levenshtein 

import time

from rapidfuzz import fuzz


from geodata import Loc, Normalize, Geodata, GeoUtil, GeoSearch


class Score:
    VERY_GOOD = 4
    STRONG_CUTOFF = VERY_GOOD + 8
    GOOD = 35
    POOR = 65
    VERY_POOR = 90


COUNTRY_IDX = 4
ADMIN1_IDX = 3
    

class MatchScore:
    """
    Calculate a heuristic score for how well a result place name matches a target place name. The score is based on percent
    of characters that didnt match plus other items - described in match_score()
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.score_diags = ''  # Diagnostic text for scoring
        self.token_weight = []
        self.prefix_weight = 0.0
        self.feature_weight = 0.0
        self.input_weight = 0.0

        # Weighting for each input term match -  adm2, adm1, country
        token_weights = [.2, .3, .5]
        self.set_weighting(token_weight=token_weights, prefix_weight=5.0, feature_weight=0.05)

        # Weighting for each part of score
        self.wildcard_penalty = 8.0

    def _calculate_wildcard_score(self, original_entry) -> float:
        if '*' in original_entry:
            # if it was a wildcard search it's hard to score - just add adjustment
            return self.wildcard_penalty
        else:
            return 0.0

    def set_weighting(self, token_weight: [], prefix_weight: float, feature_weight: float):
        """
        Set weighting of scoring components.  See match_score for details of weighting.  All weights are positive
        Args:
            token_weight:    List with Weights relative to City for County, State/Province, Country. City is 1.0   
            prefix_weight:   Weighting for prefix score
            feature_weight:  Weighting for Feature match score

        Returns:

        """
        self.token_weight = [0.0, 1.0]  # Set prefix weight to 0.0 and city to 1.0
        
        if len(token_weight) != 3:
            raise Exception('Token weight must have 3 elements: County, State/Province, Country')
        
        # Append weighting for each input term match -   adm2, adm1, country
        self.token_weight += list(token_weight)
        self.token_weight = [abs(item) for item in self.token_weight]
        self.logger.debug(f'{self.token_weight}')

        self.prefix_weight = abs(prefix_weight)
        self.feature_weight = abs(feature_weight)
        #  Feature weight must be less than 1.0.
        if self.feature_weight > 1.0:
            self.logger.error('Feature weight must be less than 1.0')
        self.input_weight = 1.0 - feature_weight

    def match_score(self, target_place: Loc, result_place: Loc) -> float:
        """
            Calculate a heuristic score for how well a result place name matches a target place name.  The score is based on
            percent of characters that didnt match in input and output (plus other items described below).
            Mismatch score is 0-100% reflecting the percent mismatch between the user input and the result.  This is then
            adjusted by Feature type (large city gives best score) plus other items to give a final heuristic where
            -10 is perfect match of a large city and 100 is no match.

            A) Heuristic:
            1) Create 5 part title (prefix, city, county, state/province, country)
            2) Normalize text - Normalize.normalize_for_scoring()
            3) Remove sequences of 2 chars or more that match in target and result
            4) Calculate inscore - percent of characters in input that didn't match result.  Weight by term (city,,county,state,ctry)
                    Exact match of city term gets a bonus
            5) Calculate result score - percent of characters in db result that didn't match input

            B) Score components (All are weighted in final score):   
            in_score - (0-100) - score for input that didnt match output   
            feature_score - (0-100)  More important features get lower score.   
            City with 1M population is zero.  Valley is 100.  Geodata.feature_priority().  
            wildcard_penalty - score is raised by X if it includes a wildcard   
            prefix_penalty -  score is raised by length of Prefix   

            C) A standard text difference, such as Levenstein, was not used because those treat both strings as equal,   
            whereas this treats the User text as more important than DB result text and also weights each token.  A user's   
            text might commonly be something like: Paris, France and a DB result of Paris, Paris, Ile De France, France.   
            The Levenstein distance would be large, but with this heuristic, the middle terms can have lower weights, and   
            having all the input matched can be weighted higher than mismatches on the county and province.  This heuristic gives   
            a score of -9 for Paris, France.   

        # Args:
            target_place:  Loc  with users entry.
            result_place:  Loc with DB result.
        # Returns:
            score
        """
        self.score_diags = ''  # Diagnostic text for scoring
        self.timing = 0
        save_prefix = target_place.prefix
        #self.logger.debug(f'pref={target_place.prefix}')

        # Create full, normalized titles (prefix,city,county,state,country)
        result_title, result_tokens, target_title, target_tokens = _prepare_input(target_place, result_place)
        #self.logger.debug(f'Res [{result_tokens}] Targ [{target_tokens}] ')
        
        # Calculate Prefix score.  Prefix is not used in search and longer is generally worse 
        prefix_score = _calculate_prefix_penalty(target_place.prefix)

        # Calculate score for  percent of input target text that matched result
        in_score = self._calculate_weighted_score(target_tokens, result_tokens)

        # Calculate score for wildcard search - wildcard searches are missing letters and need special handling
        wildcard_score = self._calculate_wildcard_score(target_place.original_entry)

        # Calculate Feature score - this ensures "important" places get higher rank (large city, etc)
        feature_score = Geodata.Geodata._feature_priority(result_place.feature)

        # Weight and add up scores - Each item is 0-100 and then weighted, except wildcard penalty
        score: float = in_score * self.input_weight + feature_score * self.feature_weight + \
                       prefix_score * self.prefix_weight + wildcard_score

        #self.logger.debug(f'SCORE {score:.1f} res=[{result_title}] pref=[{target_place.prefix}]'   
        #                  f'inSc={in_score * self.input_weight:.1f}% feat={feature_score * self.feature_weight:.1f} {result_place.feature}  '
        #                  f'wild={wildcard_score} pref={prefix_score * self.prefix_weight:.1f}')

        #self.logger.debug(self.score_diags)
        target_place.prefix = save_prefix

        return score + 8

    def _calculate_weighted_score(self, target_tokens: [], result_tokens: []) -> float:
        # Get score with tokens as is
        sc = self._weighted_score(target_tokens, result_tokens)

        # Get score with city and admin2 reversed
        sc2 = 1000.0

        # see if we get a decent score with city and admin2 reversed
        if 100 - fuzz.token_sort_ratio(target_tokens[1], result_tokens[2]) < 30:
            city = target_tokens[1]
            target_tokens[1] = target_tokens[2]
            target_tokens[2] = city
    
            sc2 = self._weighted_score(target_tokens, result_tokens) + 5.0
        return min(sc, sc2)

    def _weighted_score(self, target_tokens: [], result_tokens: []) -> float:
        num_inp_tokens = 0.0
        score = 0.0
        bonus = 0.0

        token_weight = copy.copy(self.token_weight)
        
        # Set weighting to half if target token for that term is blank
        #for idx, item in enumerate(target_tokens):
        #    if len(item) == 0:
        #        token_weight[idx] *= 0.5 
            
        #self.logger.debug(f'[{target_tokens}]  [{result_tokens}]')

        # Calculate difference each target segment to result segment (city, county, state/province, country)
        # Each segment can have a different weighting.  e.g. county can have lower weighting
        for idx, segment in enumerate(target_tokens):
            if idx < len(result_tokens) and idx > 0:
                #self.logger.debug(f'{idx}) Targ [{target_tokens[idx]}] Res [{result_tokens[idx]}]')
                
                if len(target_tokens[idx]) > 0:
                    # Calculate fuzzy Levenstein distance between words, smaller is better
                    start = time.time()
                    fz = 100.0 - fuzz.ratio(target_tokens[idx], result_tokens[idx])
                    self.timing += (time.time() - start)
                    # Calculate fuzzy Levenstein distance between Soundex of words
                    sd = 100.0 - fuzz.ratio(GeoSearch.get_soundex(target_tokens[idx]), GeoSearch.get_soundex(result_tokens[idx]))
                    value = fz * 0.6 + sd * 0.4
                    #self.logger.debug(f'    Val={value} Fuzz Text={fz:.1f} SDX={sd:.1f}')
                    # Extra bonus for good match
                    if value < 10:
                        #self.logger.debug('   Good match -10')
                        value -= 6
                    if target_tokens[idx][0:5] == result_tokens[idx][0:5]:
                        # Bonus if first letters match
                        #self.logger.debug('   First letter match -5')
                        value -= 3
                elif len(result_tokens[idx]) > 0:
                    # Target had no entry for this term
                    # Give a penalty if target didn't have country 
                    if idx == COUNTRY_IDX:
                        value = 15.0
                    elif idx == ADMIN1_IDX:
                        # Give a penalty if target didn't have state/province
                        value = 5.0
                    else:
                        value = 5.0
                    #self.logger.debug(txt)
                else:
                    if idx == COUNTRY_IDX:
                        value = 15.0
                    elif idx == ADMIN1_IDX:
                        # Give a penalty if target didn't have state/province
                        value = 5.0
                    else:
                        value = 0

                    #self.logger.debug(txt)

                    
                #if  idx >0:
                #    self.logger.debug(f'  {idx})  [{target_tokens[idx]}] [{result_tokens[idx]}]  val={value}')
                
                #if '*' in segment:
                    # Add penalty if wildcard is in segment
                    #value += 100.0
                score += value * token_weight[idx]
                num_inp_tokens += token_weight[idx]
                self.score_diags += f'  {idx}) {value} [{result_tokens[idx]}]'
            else:
                #self.logger.warning(f'Short Result len={len(result_tokens)} targ={target_tokens[idx]}')
                pass

        # Average over number of tokens (with fractional weight).  Gives 0-100 regardless of weighting and number of tokens
        if num_inp_tokens > 0:
            score = (score / num_inp_tokens) 
        else:
            score = 0
            
        return score + bonus

    @staticmethod
    def _adjust_adm_score(score, feat):
        # Currently just pass thru score
        return score


def _calculate_prefix_penalty(prefix):
    # If the location has a prefix, it is not as good a match
    prefix_len = len(prefix)
    if prefix_len > 0:
        # reduce penalty if prefix is a street (contains digits or 'street' or 'road')
        penalty = 5 + prefix_len
        if GeoUtil.is_street(prefix):
            penalty *= 0.2
    else:
        penalty =  0
    return penalty


def _prepare_input(target_place: Loc, result_place: Loc):
    # Create full, normalized  title (prefix,city,county,state,country)

    result_title = full_normalized_title(result_place)
    target_title = full_normalized_title(target_place)
    target_title, result_title = Normalize.remove_aliase(target_title, result_title)
    result_tokens = result_title.split(',')
    target_tokens = target_title.split(',')
    return result_title, result_tokens, target_title, target_tokens

def full_normalized_title(place: Loc) -> str:
    # Create a full normalized five part title (includes prefix)
    # Clean up prefix - remove any words that are in city, admin1 or admin2 from Prefix
    #place.prefix = Loc.Loc.matchscore_prefix(place.prefix, place.get_long_name(None))
    title = place.get_five_part_title()
    title = Normalize.normalize_for_scoring(title, place.country_iso)

    return title


def remove_if_input_empty(target_tokens, res_tokens):
    # Remove terms in Result if input for that term was empty
    for ix, term in enumerate(target_tokens):
        if len(term) == 0 and ix < len(res_tokens) - 1 and ix > 1:
            res_tokens[ix] = ''




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""
geoname database support routines.  Add locations to geoname DB, create geoname tables and indices.   
Provides a number of methods to lookup locations by name, feature, admin ID, etc.
"""
import copy
import logging
import re
import time
from operator import itemgetter

import phonetics

from geodata import GeoUtil, Loc, Country, MatchScore, Normalize, QueryList
from geodata.GeoUtil import Query, Result, Entry

FUZZY_LOOKUP = [Result.WILDCARD_MATCH, Result.WORD_MATCH, Result.SOUNDEX_MATCH]


class GeoSearch:
    """
    geoname database search routines.  look up items in geoname DB
    """

    def __init__(self, geodb):
        self.logger = logging.getLogger(__name__)
        self.detailed_debug = True
        self.start = 0
        self.match = MatchScore.MatchScore()
        self.use_wildcards = True
        self.total_lookups = 0
        self.cache = {}
        self.place_type = ''
        self.select_str = 'name, country, admin1_id, admin2_id, lat, lon, feature, geoid, sdx'
        self.geodb = geodb

    def lookup_place(self, place: Loc) -> []:
        """
            **Lookup a place in geoname.org db**     
            Lookup is based on place.place_type as follows:  
                Loc.PlaceType.COUNTRY: does self.search_country(place)  
                Loc.PlaceType.ADVANCED_SEARCH: does self.feature_search(place)  
                Otherwise: do self.search_city(place)  
        # Args:   
            place: Loc instance.  Call Loc.parse_place() before calling lookup_place()   
  
        # Returns:   
            None.  
            place.georow_list contains a list of matching entries.  
            Each entry has: Lat, Long, districtID (County or State or Province ID), and a match quality score  

        """
        self.start = time.time()
        place.result_type = Result.STRONG_MATCH
        best_score = MatchScore.Score.VERY_POOR

        # target_feature = place.place_type

        # Lookup Place based on Place Type
        # if place.admin1_id:
        # admin1 was already looked up 
        # self.logger.debug('admin1  already looked up')
        # return
        if place.place_type == Loc.PlaceType.COUNTRY:
            self.logger.debug('country look up')

            if place.georow_list:
                place.country_name = self.get_country_name(place.country_name, place.georow_list)
        elif place.place_type == Loc.PlaceType.ADMIN1:
            self.logger.debug('admin1 look up')
            pass
        else:
            # General search 
            # self.logger.debug('general look up')
            place.georow_list.clear()
            place.result_type = self._search(georow_list=place.georow_list, name=place.city, admin1_id=place.admin1_id,
                                             admin2_id=place.admin2_id, iso=place.country_iso, feature=place.feature, sdx=get_soundex(place.city))
            # self.logger.debug(f'lookup result {place.georow_list}')

            if len(place.georow_list) > 0:
                best_score = self.assign_scores(georow_list=place.georow_list, place=place, target_feature=place.feature,
                                                fast=False, quiet=False)
                
            if best_score >= MatchScore.Score.POOR and place.feature not in ['ADM0', 'ADM1']:
                # Get more results unless we have a high score
                # self.logger.debug('soundex/combination look up')
                row_list = []
                self.search_for_combinations(row_list, place.city, place, 'main.geodata')
                if len(row_list) > 0:
                    place.georow_list.extend(row_list)
                self.logger.debug(f'Combo result [{row_list}]')

        if len(place.georow_list) > 0:
            self.assign_scores(georow_list=place.georow_list, place=place, target_feature=place.feature,
                               fast=False, quiet=False)
            # self.logger.debug(f'Found: {len(place.georow_list)} matches  '
            #                  f' [{place.georow_list}]\n')
        else:
            # self.logger.debug(f'LOOKUP. No match:for  nm=[{place.get_five_part_title()}]\n')
            pass

    def _search(self, georow_list, name, admin1_id, admin2_id, iso, feature, sdx=''):
        """
        
        Args:
            georow_list: List of matching georows
            name: name of location to lookup
            admin1_id: admin1_id of location (if available, otherwise '')
            admin2_id: 
            iso: country iso of location (if available,  otherwise '')
            feature: feature code of location (if available, otherwise '')

        Returns: result_type

        """
        query_list = []
        # self.logger.debug(f'** SEARCH  name=[{name}] adm1 id=[{admin1_id}]'
        #                  f' adm2 id=[{admin2_id}] iso=[{iso}] feat=[{feature}] sdx=[{sdx}]')

        # Put together where clauses with all non-blank fields 
        if len(name + admin2_id + admin2_id + iso) == 0:
            return GeoUtil.Result.NO_MATCH

        ql = QueryList.QueryItem()
        where_clauses = ["name", "country", "admin1_id", "admin2_id", "feature"]
        terms = [name, iso, admin1_id, admin2_id, feature]
        ql.add_clauses(where_clauses=where_clauses, terms=terms)  # add 
        query_list.append(Query(where=ql.where, args=ql.args, result=Result.PARTIAL_MATCH))
        ql.clear()
        if len(sdx) > 0:
            where_clauses = ["country", "feature", "sdx"]
            terms = [iso, feature, sdx]
            ql.add_clauses(where_clauses=where_clauses, terms=terms)  # add 
            query_list.append(Query(where=ql.where, args=ql.args, result=Result.PARTIAL_MATCH))

        # See if we can determine feature type from name and lookup by that
        self.add_feature_query(query_list, name, iso)

        return self.geodb.process_query_list(result_list=georow_list, select_fields=self.select_str, from_tbl=ql.table,
                                             query_list=query_list)

    def get_admin2_name(self, admin1_id, admin2_id, iso) -> str:
        """
        Search for Admin2 name using admin2_id and admin1_id 

        # Args:   
            admin1_id: Admin1 ID   
            admin2_id: Admin2 ID   
            iso: country ISO     

        # Returns:
            Admin2 name
        """
        if len(admin2_id) > 0:
            return self._get_name(admin1_id=admin1_id, admin2_id=admin2_id, iso=iso, feature='ADM2')
        else:
            return ''

    def get_admin1_name(self, admin1_id, iso) -> str:
        """
        Search for Admin1 name using admin1_id (rather than place instance)

        # Args:   
            admin1_id: Admin1 ID   
            iso: country ISO     

        # Returns:
            Admin1 name.   
        """
        if len(admin1_id) > 0:
            return self._get_name(admin1_id=admin1_id, admin2_id='', iso=iso, feature='ADM1')
        else:
            return ''

    def get_country_name(self, iso: str, row_list) -> str:
        """
             return country name for specified ISO code 
        # Args:   
            iso:   Country ISO code
        # Returns:
            Country name or ''
        """
        return self._get_name(admin1_id='', admin2_id='', iso=iso, feature='ADM0')

    def _get_name(self, admin1_id, admin2_id, iso, feature, sdx='') -> str:
        """
             return  name for specified ID

        # Args:   
            iso:   Country ISO code

        # Returns:
             name or ''
        """
        row_list = []
        key = f'{admin1_id}_{admin2_id}_{iso}_{feature}'
        if len(admin1_id + admin2_id + iso) == 0:
            return ''

        name = self.cache.get(key)
        if name:
            # self.logger.debug('CACHE MATCH')
            return name

        self._search(georow_list=row_list, name='', admin1_id=admin1_id, admin2_id=admin2_id, iso=iso, feature=feature, sdx=sdx)

        if len(row_list) > 0:
            # self.logger.debug(f'GET NAME for {feature} ID: 1[{admin1_id}] 2[{admin2_id}] iso[{iso}] RES=[{row_list[0][Entry.NAME]}]')
            self.cache[key] = row_list[0][Entry.NAME]
            return row_list[0][Entry.NAME]
        else:
            # self.logger.debug(f'GET NAME for ID 1[{admin1_id}] 2[{admin2_id}] iso[{iso}] RES=[NO MATCH]')
            return ''

    def get_admin1_alternate_name(self, admin1_id, place: Loc) -> (str, str):
        """
             Get Admin1 name from alternate name table 

        # Args:   
            place:   place instance.  place.admin1_id is used for lookup

        # Returns:
            (admin1 name, lang)  place.admin2_id is updated with best match 
        """
        # self.logger.debug('GET ADMIN1 ALT NAME')

        # target = place.admin1_id
        if len(admin1_id) == 0:
            return '', ''

        # Find GEOID for admin1_id and then get alternate name for GEOID
        row_list = []
        self._search(georow_list=row_list, name='', admin1_id=admin1_id, admin2_id='', iso=place.country_iso, feature='ADM1')
        if len(row_list) > 0:
            row = row_list[0]
            # Get alternate name for this GEOID
            admin1_name, lang = self.get_alternate_name(row[Entry.ID])
            return admin1_name, lang
        else:
            return '', ''

    def get_alternate_name(self, geoid) -> (str, str):
        """
        Retrieve alternate name for specified GEOID   

        #Args:    
            geoid: Geoid to get alternate names for   

        #Returns: 
            (name, lang)   

        """
        # todo - make rowlist a param
        query_list = [
            Query(where="geoid = ?",
                  args=(geoid,),
                  result=Result.STRONG_MATCH)]
        select = 'name, lang'
        row_list = []
        self.geodb.process_query_list(result_list=row_list, select_fields=select, from_tbl='main.altname', query_list=query_list)
        if len(row_list) > 0:
            return row_list[0][0], row_list[0][1]
        else:
            return '', ''

    def update_names(self, place):
        # Use ID fields to fill in  missing names for admin1, admin2, and country
        row_list = []

        if place.admin1_id != '':
            if place.admin1_name == '':
                place.admin1_name = self.get_admin1_name(place.admin1_id, place.country_iso)
            if place.admin2_name == '':
                place.admin2_name = self.get_admin2_name(place.admin1_id, place.admin2_id, place.country_iso)
        place.country_name = str(self.get_country_name(place.country_iso, row_list))

    def get_admin1_id(self, admin1_name, place: Loc, row_list):
        """
        Search for Admin1 ID using admin1_name

        # Args:   
            admin1_name:
            place:   

        # Returns:
            None.  place.admin1_id and place.country_iso are updated with best match 
        """

        # TODO get rid of utilization of place, return id
        # if len(admin1_name) == 0:
        #    return
        admin1_name = Normalize.admin1_normalize(admin1_name, place.country_iso)
        # self.logger.debug(f'GET ADMIN1 ID from [{admin1_name}]')

        result_place = copy.copy(place)
        sdx = get_soundex(admin1_name)

        self._search(georow_list=row_list, name=admin1_name, admin1_id='', admin2_id='', iso=place.country_iso, feature='ADM1', sdx=sdx)

        # Sort places in match_score order
        if len(row_list) == 0:
            self.logger.debug(f'search for combo {admin1_name}')
            self.search_for_combinations(row_list=row_list, target=admin1_name, place=place, table='main.admin')
        else:
            # self.logger.debug(f'found {row_list}')
            pass

        if len(row_list) > 0:
            self.assign_scores(row_list, result_place, 'ADM1', fast=False, quiet=False)
            sorted_list = sorted(row_list, key=itemgetter(GeoUtil.Entry.SCORE))
            place.admin1_id = sorted_list[0][Entry.ADM1]

            # self.logger.debug(f'SEARCH for ADMIN1 ID DONE -- Found adm1 id = [{place.admin1_id}] row={sorted_list[0]} ')
            # Fill in Country ISO
            if place.country_iso == '':
                place.country_iso = sorted_list[0][Entry.ISO]
        else:
            self.logger.debug(f'SEARCH ADMIN1 ID DONE -- NO MATCH SEARCH  adm1 name = {admin1_name}  ')

    def get_country_iso(self, georow_list, country_name, place: Loc) -> str:
        """
             return country ISO code for place.country_name   

        # Args:   
            place:   place instance.  looks up by place.country_name   

        # Returns:   
            Country ISO or ''.  If found, update place.country_name with DB country name   
        """
        # self.logger.debug(f'GET COUNTRY ISO for [{country_name}]')

        country_name, modified = Normalize.country_normalize(country_name)
        if len(country_name) == 0:
            return ''
        sdx = get_soundex(country_name) + '*'

        place.result_type = self._search(georow_list=georow_list, name=country_name, admin1_id='', admin2_id='', iso='', feature='ADM0', sdx='')
        self.assign_scores(georow_list, place, 'ADM0', fast=False, quiet=False)
        #if len(georow_list) > 0:
        #    self.logger.debug(georow_list[0])

        if place.result_type == Result.STRONG_MATCH:
            iso = georow_list[0][Entry.ISO]
            place.country_name = georow_list[0][Entry.NAME]
        else:
            place.result_type = self._search(georow_list=georow_list, name='', admin1_id='', admin2_id='', iso='', feature='ADM0',
                                             sdx=sdx)
            self.assign_scores(georow_list, place, 'ADM0', fast=False, quiet=False)
            if place.result_type == Result.STRONG_MATCH:
                iso = georow_list[0][Entry.ISO]
                place.country_name = georow_list[0][Entry.NAME]
            else:
                iso = ''

        # self.logger.debug(f'found iso [{iso}]')

        return iso

    def search_for_combinations(self, row_list, target, place, table):
        """
        Search for soundex of combinations of words in target
        Words are sorted and then Soundex searched with every combination of one word missing
        Args:
            row_list: 
            target: 
            place: 
            table: 

        Returns:

        """
        query_list = []
        sdx = get_soundex(target)

        if len(sdx) > 3 and len(place.country_iso) > 0:
            # append wildcard
            sdx += '%'

            sdx_word_list = sdx.split(' ')

            if len(sdx_word_list) == 1:
                #  one word - lookup as is
                if place.feature:
                    query_list.append(Query(where="sdx like ? AND country = ? AND feature = ?",
                                            args=(sdx, place.country_iso, place.feature,),
                                            result=Result.SOUNDEX_MATCH))
                else:
                    query_list.append(Query(where="sdx like ? AND country = ?",
                                            args=(sdx, place.country_iso,),
                                            result=Result.SOUNDEX_MATCH))
            else:
                # Multiple words - Try every combination with a single word removed
                for ignore_word_idx in range(0, len(sdx_word_list)):
                    pattern = ''
                    for idx, word in enumerate(sdx_word_list):
                        if idx != ignore_word_idx:
                            if len(pattern) == 0:
                                pattern += word
                            else:
                                pattern += ' ' + word
                    # self.logger.debug(f'{ignore_word_idx}) {pattern}')
                    if place.feature:
                        query_list.append(Query(where="sdx like ? AND country = ? AND feature = ?",
                                                args=(pattern, place.country_iso, place.feature,),
                                                result=Result.SOUNDEX_MATCH))
                    else:
                        query_list.append(Query(where="sdx like ? AND country = ?",
                                                args=(pattern, place.country_iso,),
                                                result=Result.SOUNDEX_MATCH))

                # Also try search for soundex of just first term
                pattern = sdx_word_list[0] + '%'
                query_list.append(Query(where="sdx like ? AND country = ?",
                                        args=(pattern, place.country_iso,),
                                        result=Result.SOUNDEX_MATCH))

            #for qr in query_list:
            #    self.logger.debug(f'where {qr.where} arg={qr.args} ')
                
            place.result_type = self.geodb.process_query_list(result_list=row_list, select_fields=self.select_str,
                                                              from_tbl=table, query_list=query_list, debug=True)

    def lookup_geoid(self, georow_list, geoid, place: Loc, admin=False) -> None:
        """
        Search by GEOID 
        Args:
            georow_list: 
            geoid: 
            place: 
            admin: If True, look up in admin table, otherwise look in geodata table

        Returns: None.  georow_list contains list of matches
        """
        if admin:
            table = 'main.admin'
        else:
            table = 'main.geodata'
        query_list = [
            Query(where="geoid = ? ",
                  args=(geoid,),
                  result=Result.STRONG_MATCH)
            ]
        self.geodb.process_query_list(result_list=georow_list, select_fields=self.select_str,
                                      from_tbl=table, query_list=query_list)
        if georow_list:
            # self.logger.debug(georow_list[0])
            self.assign_scores(georow_list=place.georow_list, place=place, target_feature=place.feature,
                               fast=False, quiet=False)
            # self.copy_georow_to_place(georow_list[0], place, fast=True)

    def lookup_dbid(self, georow_list, dbid, place: Loc, admin=False) -> None:
        """
        Search by Database ID 
        Args:
            georow_list: 
            dbid: 
            place: 
            admin: If True, look up in admin table, otherwise look in geodata table

        Returns: None.  georow_list contains list of matches
        """
        if admin:
            table = 'main.admin'
        else:
            table = 'main.geodata'
        query_list = [
            Query(where="id = ? ",
                  args=(dbid,),
                  result=Result.STRONG_MATCH)
            ]
        self.geodb.process_query_list(result_list=georow_list, select_fields=self.select_str,
                                      from_tbl=table, query_list=query_list)
        if georow_list:
            # self.logger.debug(georow_list[0])
            self.copy_georow_to_place(georow_list[0], place, fast=True)
        else:
            self.logger.debug('no match')

    def copy_georow_to_place(self, row, place: Loc, fast: bool):
        """
        Copy data from DB row into place instance   
        Country, admin1_id, admin2_id, city, lat/lon, feature, geoid are updated if available   
        #Args:   
            row: georow from geoname database   
            place: Loc instance   
            fast: Currently ignored
        #Returns:   
            None.  Place instance is updated with data from georow   
        """
        place.admin1_id = ''
        place.admin2_id = ''
        place.admin1_name = ''
        place.admin2_name = ''
        place.city = ''

        place.country_iso = str(row[Entry.ISO])
        row_list = []
        place.lat = row[Entry.LAT]
        place.lon = row[Entry.LON]
        place.feature = str(row[Entry.FEAT])
        #self.logger.debug(f'feat={place.feature}')
        place.geoid = str(row[Entry.ID])
        place.prefix = row[Entry.PREFIX]
        place.place_type = Loc.PlaceType.CITY

        if place.feature == 'ADM0':
            place.place_type = Loc.PlaceType.COUNTRY
            pass
        elif place.feature == 'ADM1':
            place.admin1_id = row[Entry.ADM1]
            place.place_type = Loc.PlaceType.ADMIN1
        elif place.feature == 'ADM2':
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.place_type = Loc.PlaceType.ADMIN2
        else:
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.city = row[Entry.NAME]

        if place.admin1_id != '':
            if place.admin1_name == '':
                place.admin1_name = self.get_admin1_name(place.admin1_id, place.country_iso)
            if place.admin2_name == '':
                place.admin2_name = self.get_admin2_name(place.admin1_id, place.admin2_id, place.country_iso)
        place.country_name = str(self.get_country_name(row[Entry.ISO], row_list))

        if place.admin2_name is None:
            place.admin2_name = ''
        if place.admin1_name is None:
            place.admin1_name = ''

        place.city = str(place.city)
        if place.city is None:
            place.city = ''

        try:
            place.score = row[Entry.SCORE]
        except IndexError:
            pass

        #self.logger.debug(f'<<<<< COPY DONE:  A1 [{place.admin1_name}] A2 [{place.admin2_name}] cty [{place.country_name}]'
        #                  f'lat [{place.lat}]')

    @staticmethod
    def create_wildcard(pattern):
        """
        Create wildcard pattern.  Convert * to %.  Add % on end   
        #Args:   
            pattern:   

        #Returns: wildcard pattern   

        """
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'%{pattern}%'

    @staticmethod
    def create_admin1_wildcard(pattern):
        """
        Create wildcard pattern.  Convert * to %.  Add % on end.  Currently this  
        is the same at create_wildcard   
        #Args:   
            pattern:    

        #Returns: wildcard pattern   

        """
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'{pattern}%'

    @staticmethod
    def create_county_wildcard(pattern):
        """
        create wildcard for county lookup  
        #Args:   
            pattern:    

        #Returns:  

        """
        # Try pattern with 'shire' removed
        pattern = re.sub(r"shire", "", pattern)
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'{pattern}'

    def set_display_names(self, temp_place):
        """
            See if there is an alternate name entry for this place   

        #Args:   
            temp_place: place instance   

        #Returns: None   

        """
        place_lang = Country.Country.get_lang(temp_place.country_iso)
        res, lang = self.get_alternate_name(temp_place.geoid)
        if res != '':  # and (lang == place_lang or lang == 'ut8'):
            temp_place.city = res

        res, lang = self.get_admin1_alternate_name(temp_place.city, temp_place)
        if res != '':  # and (lang == place_lang or lang == 'ut8'):
            temp_place.admin1_name = res

    def debugZZZ(self, text):
        if self.detailed_debug:
            self.logger.debug(text)

    def add_feature_query(self, query_list, target, iso):
        # Scan target to see if we can determine what feature type it is
        word, group = GeoUtil.get_feature_group(target)
        if word != '':
            if 'st ' not in word:
                targ = '%' + re.sub(word, '', target).strip(' ') + '%'
            else:
                targ = target
            query_list.append(Query(where="name LIKE ? AND country = ? AND feature like ?",
                                    args=(targ, iso, group),
                                    result=Result.WORD_MATCH))
            self.logger.debug(f'Add Feature query [{targ}] {group}')

    def assign_scores(self, georow_list, place, target_feature, fast=False, quiet=False) -> float:
        """
                    Assign match score to each result in list   
        Args:
            place: 
            target_feature: 
            fast: 
            quiet: if True, set logging to INFO

        Returns:

        """
        result_place: Loc = Loc.Loc()

        min_score = 9999
        original_prefix = place.prefix

        # If quiet, then only log at INFO level
        lev = logging.getLogger().getEffectiveLevel()
        if quiet:
            logging.getLogger().setLevel(logging.INFO)

        # Add match quality score and prefix to each entry
        for idx, rw in enumerate(georow_list):
            place.prefix = original_prefix
            if len(rw) == 0:
                continue
            # self.logger.debug(rw)
            self.copy_georow_to_place(row=rw, place=result_place, fast=fast)
            result_place.original_entry = result_place.get_long_name(None)
            #self.logger.debug(f'plac feat=[{place.feature}] targ=[{target_feature}]')
            if result_place.feature == target_feature:
                bonus = 10.0
            else:
                bonus = 0

            if len(place.prefix) > 0 and result_place.prefix == '':
                result_place.prefix = ' '
                # result_place.prefix_commas = ','
            else:
                result_place.prefix = ''

            # Remove items in prefix that are in result
            if place.place_type != Loc.PlaceType.ADVANCED_SEARCH:
                result_name = result_place.get_long_name(None)
                place.prefix = Loc.Loc.prefix_cleanup(place.prefix, result_name)
            else:
                place.updated_entry = place.get_long_name(None)

            score = self.match.match_score(target_place=place, result_place=result_place) - bonus
            min_score = min(min_score, score)

            # Convert row tuple to list and extend so we can assign score
            update = list(rw)
            if len(update) < GeoUtil.Entry.SCORE + 1:
                update.append(1)
            update[GeoUtil.Entry.SCORE] = score

            result_place.prefix = Normalize.normalize(place.prefix, True)
            update[GeoUtil.Entry.PREFIX] = result_place.prefix
            georow_list[idx] = tuple(update)  # Convert back from list to tuple
            # self.logger.debug(f'{update[GeoUtil.Entry.SCORE]:.1f} {update[GeoUtil.Entry.NAME]} [{update[GeoUtil.Entry.PREFIX]}]')

        #if len(georow_list) > 0:
        #    self.logger.debug(f'min={min_score} {georow_list[0]}')
        if min_score < MatchScore.Score.STRONG_CUTOFF:
            place.result_type = GeoUtil.Result.STRONG_MATCH

        # Restore logging level
        logging.getLogger().setLevel(lev)
        return min_score


def special_terms_last(item):
    """  key function - return key with special terms forced to sort last"""
    special = {'east': 1, 'west': 1, 'north': 1, 'south': 1}
    if special.get(item):
        # Force special term to sort last.  { will sort last
        return '{' + item
    return item


def get_soundex(text) -> str:
    """
    Returns: Phonetics Double Metaphone Soundex code for sorted words in text  
    First two actual letters of word are prepended
    """
    sdx = []
    # Remove l' and d' from soundex 
    text = re.sub(r"l\'", "", text)
    text = re.sub(r"d\'", "", text)

    # Force special terms to sort at end of list so wildcard will find phrases without them  (East, West, North, South)
    # North Haden becomes Haden North and Haden% will match it.
    word_list = sorted(text.split(' '), key=special_terms_last)

    for word in word_list:
        if len(word) > 1:
            sdx.append(word[0:2] + phonetics.dmetaphone(word)[0])
        else:
            sdx.append(phonetics.dmetaphone(word)[0])
    res = ' '.join(sdx)
    if len(res) == 0:
        res = text
    return res.lower()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
""" Utilities """
import collections
import logging
import os
import re
import sys
from difflib import SequenceMatcher


class Entry:
    # Database geodata table and admin table entries
    NAME = 0
    ISO = 1
    ADM1 = 2
    ADM2 = 3
    LAT = 4
    LON = 5
    FEAT = 6
    ID = 7
    SDX = 8
    PREFIX = 8  # Note - item 8 is overloaded:  Soundex in DB and Prefix in result
    SCORE = 9
    MAX = 9


class Result:
    # Result codes for lookup
    STRONG_MATCH = 9
    MULTIPLE_MATCHES = 8
    PARTIAL_MATCH = 7
    WORD_MATCH = 6
    WILDCARD_MATCH = 5
    SOUNDEX_MATCH = 4
    DELETE = 3
    NO_COUNTRY = 2
    NO_MATCH = 1
    NOT_SUPPORTED = 0


# Result types that are successful matches
successful_match = [Result.STRONG_MATCH, Result.PARTIAL_MATCH, Result.WILDCARD_MATCH, Result.WORD_MATCH,
                    Result.SOUNDEX_MATCH, Result.MULTIPLE_MATCHES]

Query = collections.namedtuple('Query', 'where args result')


def get_directory_name() -> str:
    """
    Returns: Name of geodata data directory where geonames.org files are
    """
    return "geoname_data"


def get_cache_directory(basepath):
    """ 
    Returns:  directory for geodata cache files including DB
    """
    return os.path.join(basepath, "cache")


def is_street(text) -> bool:
    # See if text looks like a street name
    street_patterns = [r'\d', ' street', ' avenue', ' road', 'rue ']
    for pattern in street_patterns:
        if bool(re.search(pattern, text)):
            return True
    return False


def set_debug_logging(msg):
    """
         Set up logging configuration for debug level 
    # Args:
        msg: Initial message to log
    """
    logger = logging.getLogger(__name__)
    fmt = "%(levelname)s %(name)s.%(funcName)s %(lineno)d: %(message)s"
    logging.basicConfig(level=logging.DEBUG, stream=sys.stdout, format=fmt)
    logger.info(msg)
    return logger


def set_info_logging(msg):
    """
         Set up logging configuration for info level 
    # Args:
        msg: Initial message to log
    """
    logger = logging.getLogger(__name__)
    fmt = "%(levelname)s %(name)s.%(funcName)s %(lineno)d: %(message)s"
    logging.basicConfig(level=logging.INFO, stream=sys.stdout, format=fmt)
    logger.info(msg)
    return logger


def _remove_matching_seq(text1: str, text2: str, attempts: int, min_len: int) -> (str, str):
    """
    Find largest matching sequence.  Remove it in text1 and text2.
            called by remove_matching_sequences which provides a wrapper
    Call recursively until attempts hits zero or there are no matches longer than 1 char
    :param text1:
    :param text2:
    :param attempts: Number of times to remove largest text sequence
    :return:
    """
    s = SequenceMatcher(None, text1, text2)
    match = s.find_longest_match(0, len(text1), 0, len(text2))
    if match.size >= min_len:
        # Remove matched sequence from inp and out
        item = text1[match.a:match.a + match.size]
        text2 = re.sub(item, '', text2, count=4)
        text1 = re.sub(item, '', text1, count=4)
        if attempts > 0:
            # Call recursively - get next largest match and remove it
            text1, text2 = _remove_matching_seq(text1, text2, attempts - 1, min_len)
    return text1, text2


def remove_matching_sequences(text1: str, text2: str, min_len: int) -> (str, str):
    """
    Find largest sequences that match between text1 and 2.  Remove them from text1 and text2.
    Matches will NOT include commas
    # Args:
        text1:
        text2:
        min_len: minimum length of match that will be removed
    Returns: text 1 and 2 with the largest text sequences in both removed
    """
    # Prepare strings for input to remove_matching_seq
    # Swap all commas in text2 string to '@'.  This way they will never match comma in text1 string
    # Ensures we don;t remove commas and don't match across tokens
    text2 = re.sub(',', '@', text2)
    text1, text2 = _remove_matching_seq(text1=text1, text2=text2, attempts=15, min_len=min_len)
    # Restore commas in text2
    text2 = re.sub('@', ',', text2)
    return text1.strip(' '), text2.strip(' ')


def _lowercase_match_group(matchobj):
    return matchobj.group().lower()


def capwords(text):
    """
    Change text to Title Case. Fixes the apostrophe handling problem with title() 
    """
    if text is not None:
        # Fix handling for contractions not handled correctly by title()
        poss_regex = r"(?<=[a-z])[\']([A-Z])"
        text = re.sub(poss_regex, _lowercase_match_group, text.title())

    return text


def get_feature_group(location: str):
    """
        Scan location name to see if it refers to a generic feature like cemetery or church,etc.
        This allows a database search by geonames.org feature type
        If any word in location is in feature list, return the word found and the feature group 
    Args:
        location: 

    Returns: (word, group) - feature word found, geonames.org feature group it is in

    """
    for word in location.split(' '):
        group = feature_mapping.get(word)
        if group:
            return word, group
        
    return '', ''

feature_list = ["ADM1", "ADM2", "ADM3", "ADM4", "ADMF", "AREA", "CH", "CSTL", "CMTY", "EST ", "HSP", "FT",
           "HSTS", "ISL", "MSQE", "MSTY", "MT", "MUS", "PAL", "PPL", "PPLA", "PPLA2", "PPLA3", "PPLA4",
           "PPLC", "PPLG", "PPLH", "PPLL", "PPLQ", "PPLX", "PRK", "PRN", "PRSH", "RUIN", "RGN", "RLG", "STG",
           "SQR", "SYG", "VAL", "MNMT"]

feature_mapping = {
    # If location has item below, then derive its feature type
    'st'           : 'CH',
    'kapelle'      : 'CH',
    'chapel'       : 'CH',
    'kirche'       : 'CH',
    'kirke'        : 'CH',
    'church'       : 'CH',
    'eglise'       : 'CH',
    'cathedral'    : 'CH',
    'cathedrale'   : 'CH',
    'kathedrale'   : 'CH',
    'kanisat'      : 'CH',
    'iglesia'      : 'CH',
    'catedral'     : 'CH',
    'parroquia'    : 'CH',
    'chapelle'     : 'CH',
    'chiesa'       : 'CH',
    'cappella'     : 'CH',
    'basilica'     : 'CH',
    'cattedrale'   : 'CH',

    'castle'       : 'CSTL%',
    'schloss'      : 'CSTL%',
    'chateau'      : 'CSTL%',
    'castell'      : 'CSTL%',
    'castillo'     : 'CSTL%',
    'castello'     : 'CSTL%',
    'kasteel'      : 'CSTL%',
    'zamek'        : 'CSTL%',
    'slott'        : 'CSTL%',

    'cemetery'     : 'CMTY',
    'cimetiere'    : 'CMTY',
    'cementerio'   : 'CMTY',
    'cimitero'     : 'CMTY',
    'rooms katholieke begraafplaats': 'CMTY',
    'begraafplaats': 'CMTY',
    'gravlund'     : 'CMTY',
    'kirkegard'    : 'CMTY',
    'friedhof'     : 'CMTY',

    'hospital'     : 'HSP',
    'hopital'      : 'HSP',
    'krankenhaus'  : 'HSP',
    'ospedale'     : 'HSP',

    'palace'        : 'PAL',
    'palais': 'PAL',

    'county'       : 'ADM2',
    'island'       : 'ISL',

    'mountain'     : "MT",
    'museum'       : "MUS",
    'mosque'       : "MSQE",
    'monastery'    : "MSTY",
    'park'         : "PRK",
    'prison'       : "PRN",
    'square'       : "SQR",
    'synagogue'    : "SYG",
    'abbey'         : "MSTY%",
    'priory'        : "MSTY%",
    }

feature_names = {
    "AREA" : 'Area',
    "CH"   : 'Church',
    "CSTL" : 'Castle',
    "CMTY" : 'Cemetery',
    "EST"  : 'Estate',
    "HSP"  : 'Hospital',
    "HSTS" : 'Historical',
    "ISL"  : 'Island',
    "MT"   : 'Mountain',
    "MUS"  : 'Museum',
    "MSQE" : 'Mosque',
    "MSTY" : 'Monastery',
    "PAL"  : 'Palace',
    "PRK"  : 'Park',
    "PRN"  : 'Prison',
    "RUIN" : 'Ruin',
    "SQR"  : 'Square',
    "VAL"  : 'Valley',
    "ADM1" : 'State',
    "ADM2" : 'County',
    "ADM3" : 'Township',
    "ADM4" : 'Township',
    "PPL"  : 'City',
    "PPLA" : 'City',
    "PPLA2": 'City',
    "PPLA3": 'City',
    "PPLA4": 'City',
    "PPLC" : 'City',
    "PPLG" : 'City',
    "PPLH" : 'City',
    "PPLL" : 'Village',
    "PPLQ" : 'Historical',
    "PPLX" : 'Neighborhood',
    "SYG"  : 'Synagogue'
    }
